//=============================================================================
//
//  CLASS CSurfacePlane - IMPLEMENTATION
//
//=============================================================================
#define RAYTRACER_SURFACEPLANE_CPP

//== INCLUDES =================================================================
#include "Plane.h"


//== IMPLEMENTATION ==========================================================
void Plane::renderGL() const
{

    //Generate plane
    glPushMatrix();
    if( m_selected )
    {
        GLfloat mat_ambient[] = { 1, 1, 1 };
        glMaterialfv( GL_FRONT, GL_AMBIENT, mat_ambient );
    }
    else
        m_shader->shadeGL();

    VectorType3 nx( 1, 0, 0 );
    VectorType3 nz( 0, 0, 1 );

    //Rotation to xy-plane
    float       dx = dot( nz, m_v3Normal ) / m_v3Normal.length();
    float       dz = dot( nx, m_v3Normal ) / m_v3Normal.length();
    dx = 90 - 180 * acos( dx ) / M_PI;
    dz = 90 - 180 * acos( dz ) / M_PI;
    glRotatef( dx, 1, 0, 0 );
    glRotatef( dz, 0, 0, 1 );
    glTranslatef( m_v3Point[0], m_v3Point[1], m_v3Point[2] );
    glBegin( GL_QUADS );    // rendering the plane quad. Note, it should be big

    // enough to cover all clip edge area.
    glVertex3f( -10000, 0, -10000 );
    glVertex3f( -10000, 0, 10000 );
    glVertex3f( 10000, 0, 10000 );
    glVertex3f( 10000, 0, -10000 );
    glNormal3f( m_v3Normal[0], m_v3Normal[1], m_v3Normal[2] );
    glEnd();
    glPopMatrix();
}

//translate object
void Plane::translate(
    float   x,  
    float   y,  
    float   z ) 
{
    m_v3Point[0] = x;
    m_v3Point[1] = y;
    m_v3Point[2] = z;
}

//rotate object
void Plane::rotate(
    float   rotx,   
    float   /*roty*/,   
    float   rotz )  
{

    //rot y ohne Funktion bei Ebenen
    //Drehmatrix um x-Achse
    rotx = ( rotx ) * M_PI / 180;
    rotz = ( rotz ) * M_PI / 180;

    VectorType3     x1 = VectorType3( 1, 0, 0 );
    VectorType3     x2 = VectorType3( 0, cos( rotx ), -sin( rotx ) );
    VectorType3     x3 = VectorType3( 0, sin( rotx ), cos( rotx ) );
    MatrixType33    rx;
    rx.setRow( 0, x1 );
    rx.setRow( 1, x2 );
    rx.setRow( 2, x3 );

    VectorType3     nx( 1, 0, 0 );
    VectorType3     nz( 0, 0, 1 );
    VectorType3     z1 = VectorType3( cos( rotz ), -sin( rotz ), 0 );
    VectorType3     z2 = VectorType3( sin( rotz ), cos( rotz ), 0 );
    VectorType3     z3 = VectorType3( 0, 0, 1 );
    MatrixType33    rz;
    rz.setRow( 0, z1 );
    rz.setRow( 1, z2 );
    rz.setRow( 2, z3 );

    VectorType3 ny = VectorType3( 0, 1, 0 );
    m_v3Normal = rx * rz * ny;
}


void Plane::getRotation(
    int *rotx,  
    int *roty,  
    int *rotz ) 
{
    VectorType3 nx( 1, 0, 0 );
    VectorType3 nz( 0, 0, 1 );

    //Rotation to xy-plane
    *rotx = 90 - 180 * acos( dot( nz, m_v3Normal ) / m_v3Normal.length() ) / M_PI;
    *rotz = 90 - 180 * acos( dot( nx, m_v3Normal ) / m_v3Normal.length() ) / M_PI;
    *roty = 0;
}


void Plane::constructONB(
    const VectorType3   &v, 
    VectorType3         &x, 
    VectorType3         &y, 
    VectorType3         &z )  const
{
    z = v;
    z.normalize();

    //construct an orthogonal vector
    if( fabs( z[0] ) < fabs( z[1] ) && fabs( z[0] ) < fabs( z[2] ) )
    {
        x[0] = 0;
        x[1] = z[2];
        x[2] = -z[1];
    }
    else if( fabs( z[1] ) < fabs( z[2] ) )
    {
        x[0] = z[2];
        x[1] = 0;
        x[2] = -z[0];
    }
    else
    {
        x[0] = z[1];
        x[1] = -z[0];
        x[2] = 0;
    }

    x.normalize();
    y = x % z;
}

//-----------------------------------------------------------------------------
bool Plane::intersect(
    const Ray       &clRay, 
    double        t0,     
    double        t1,     
    TracingContext  &tContext )  const
{
    double    f_denominator = m_v3Normal | clRay.getDir();
    if( fabs( f_denominator ) < FLT_EPSILON )
        return false;

    double    t = ( m_v3Normal | (m_v3Point - clRay.getOrigin()) ) /
        f_denominator;
    if( t > t0 && t < t1 )
    {
        tContext.m_t = t;
        tContext.m_v3HitPoint = clRay.evaluate( t );
        tContext.m_v3Normal = m_v3Normal;

        VectorType3 v3_in_plane = tContext.m_v3HitPoint - m_v3Point;
        tContext.m_pclShader = getShader();
        return true;
    }

    return false;
}
