//=============================================================================
//
//  Type Definitions
//
//=============================================================================
#ifndef RAYTRACER_TYPES_H
#define RAYTRACER_TYPES_H

//== INCLUDES =================================================================
#include "Matrix33T.h"
#include "ColorT.h"
#include <vector>

//== NAMESPACE ================================================================
#ifndef M_PI
#define M_PI    3.141592653589793238
#endif

//== FORWARD REFS =============================================================
class                           Shader;

//== TYPEDEFS =========================================================
typedef VectorT<double, 4>    VectorType4;
typedef VectorT<double, 3>    VectorType3;
typedef VectorT<double, 2>    VectorType2;
typedef Color3T<double>       ColorType;
typedef Matrix33T<double>     MatrixType33;
typedef std::vector<double>   RealArrayType;

/*! Needed for Ray-Visualization in OpenGL
    Holds start- and endpoint of a ray */
struct GLRay
{
    VectorType3 m_start;
    VectorType3 m_stop;
};
typedef std::vector<GLRay>  RayArray;

/*!
    Holds all information concerning intersection
  */
struct TracingContext
{

    /*!
      Ray parameter at intersection point
      Filled by CSurface.Intersect() 
    */
    double            m_t;

    /*!
      3D-coordinate of intersection point.
      Filled by CSurface.Intersect()
    */
    VectorType3         m_v3HitPoint;

    /*!
      Normal at intersection point
      Filled by CSurface.Intersect()
    */
    VectorType3         m_v3Normal;

    /*!
      Incoming (light) direction at intersection point
      Filled by CLight.ShadowFeeler()
    */
    VectorType3         m_v3Incoming;

    /*!
      outgoing (view) direction at intersection point
      Filled by CWhitted.TraceRay()
    */
    VectorType3         m_v3Outgoing;

    /*!
      Rgb-power of light at intersection point
    */
    ColorType           m_colLightIntensity;

    /*! Reference to shader at intersection point
      Filled by CSurface.Intersect()
    */
    const Shader        *m_pclShader;

    /*! Array of Rays: Needed only for OpenGL-Visualization */
    std::vector<GLRay>  m_glRays;
};
static const double   SELF_ISECT_OFFSET = 1.0e-3;
static const double   EPSILON = 1.0e-6;



//Define which Acceleration Type to use
enum RenderingMode
{
      MODE_DISPLAY,
      MODE_TIMING
};

//=============================================================================
#endif // RAYTRACER_TYPES_H
